local VERSION = tonumber(("$Revision: 44289 $"):match("%d+"))

local Parrot = Parrot
local Parrot_Display = Parrot:NewModule("Display", "AceEvent-2.0", "AceHook-2.1")
if Parrot.revision < VERSION then
	Parrot.version = "r" .. VERSION
	Parrot.revision = VERSION
	Parrot.date = ("$Date: 2007-07-19 23:29:38 -0400 (Thu, 19 Jul 2007) $"):match("%d%d%d%d%-%d%d%-%d%d")
end

local L = Parrot:L("Parrot_Display")

local SharedMedia = AceLibrary("SharedMedia-1.0")

local new, del = Parrot.new, Parrot.del

local ParrotFrame

Parrot_Display.db = Parrot:AcquireDBNamespace("Display")
Parrot:RegisterDefaults("Display", 'profile', {
	font = "Friz Quadrata TT",
	fontSize = 18,
	fontOutline = "THICKOUTLINE",
	stickyFont = "Friz Quadrata TT",
	stickyFontSize = 26,
	stickyFontOutline = "THICKOUTLINE",
})

local Parrot_AnimationStyles
local Parrot_Suppressions
local Parrot_ScrollAreas
function Parrot_Display:OnEnable()
	Parrot_AnimationStyles = Parrot:GetModule("AnimationStyles")
	Parrot_Suppressions = Parrot:GetModule("Suppressions")
	Parrot_ScrollAreas = Parrot:GetModule("ScrollAreas")
	if not ParrotFrame then
		ParrotFrame = CreateFrame("Frame", "ParrotFrame", UIParent)
		ParrotFrame:SetFrameStrata("HIGH")
		ParrotFrame:SetToplevel(true)
		ParrotFrame:SetPoint("CENTER")
		ParrotFrame:SetWidth(0.0001)
		ParrotFrame:SetHeight(0.0001)
	end
	
	if _G.CombatText_AddMessage then
		self:Hook("CombatText_AddMessage", true)
	else
		function _G.CombatText_AddMessage(...)
			self:CombatText_AddMessage(...)
		end
	end
end

function Parrot_Display:OnDisable()
end

function Parrot_Display:CombatText_AddMessage(message, scrollFunction, r, g, b, displayType, isStaggered)
	if type(message) ~= "string" then
		return
	end
	if type(r) ~= "number" or type(g) ~= "number" or type(b) ~= "number" then
		r, g, b = 1, 1, 1
	end
	self:ShowMessage(message, "Notification", displayType == "crit", ("%02x%02x%02x"):format(r*255, g*255, b*255), nil, nil, nil)
end

function Parrot_Display:OnOptionsCreate()
	local outlineValidate = {
		NONE = L["None"],
		OUTLINE = L["Thin"],
		THICKOUTLINE = L["Thick"],
	}

	Parrot.options.args.general.args.font = {
		type = 'group',
		name = L["Master font settings"],
		desc = L["Master font settings"],
		args = {
			normalFace = {
				type = 'text',
				name = L["Normal font"],
				desc = L["Normal font face."],
				get = function()
					return Parrot_Display.db.profile.font
				end,
				set = function(value)
					Parrot_Display.db.profile.font = value
				end,
				validate = SharedMedia:List("font"),
			},
			normalSize = {
				type = 'range',
				name = L["Normal font size"],
				desc = L["Normal font size"],
				min = 12,
				max = 32,
				step = 1,
				get = function()
					return Parrot_Display.db.profile.fontSize
				end,
				set = function(value)
					Parrot_Display.db.profile.fontSize = value
				end,
			},
			normalBorder = {
				type = 'text',
				name = L["Normal outline"],
				desc = L["Normal outline"],
				validate = outlineValidate,
				get = function()
					return Parrot_Display.db.profile.fontOutline
				end,
				set = function(value)
					Parrot_Display.db.profile.fontOutline = value
				end,
			},
			stickyFace = {
				type = 'text',
				name = L["Sticky font"],
				desc = L["Sticky font face."],
				get = function()
					return Parrot_Display.db.profile.stickyFont
				end,
				set = function(value)
					Parrot_Display.db.profile.stickyFont = value
				end,
				validate = SharedMedia:List("font"),
			},
			stickySize = {
				type = 'range',
				name = L["Sticky font size"],
				desc = L["Sticky font size"],
				min = 12,
				max = 32,
				step = 1,
				get = function()
					return Parrot_Display.db.profile.stickyFontSize
				end,
				set = function(value)
					Parrot_Display.db.profile.stickyFontSize = value
				end,
			},
			stickyBorder = {
				type = 'text',
				name = L["Sticky outline"],
				desc = L["Sticky outline"],
				validate = outlineValidate,
				get = function()
					return Parrot_Display.db.profile.stickyFontOutline
				end,
				set = function(value)
					Parrot_Display.db.profile.stickyFontOutline = value
				end,
			},
		}
	}
end

local freeFontStrings = {}
local wildFontStrings
local fontString_num = 0
local fontStringIDs = {}

function Parrot_Display:ShowMessage(text, scrollArea, sticky, color, font, fontSize, outline)
	self = Parrot_Display -- for those who do Parrot:ShowMessage
	if not Parrot:IsModuleActive(self) then
		return
	end
	if Parrot_Suppressions:ShouldSuppress(text) then
		return
	end
	if scrollArea == nil and sticky == nil then
		sticky = true
	end
	if not Parrot_ScrollAreas:HasScrollArea(scrollArea) then
		if Parrot_ScrollAreas:HasScrollArea("Notification") then
			scrollArea = "Notification"
		else
			scrollArea = Parrot_ScrollAreas:GetRandomScrollArea()
			if not scrollArea then
				return
			end
		end
	end
	local fs = next(freeFontStrings)
	if fs then
		fs:SetAlpha(1)
		fs:ClearAllPoints()
		freeFontStrings[fs] = nil
	else
		fontString_num = fontString_num + 1
		fs = ParrotFrame:CreateFontString("ParrotFrameFontString" .. fontString_num, "ARTWORK", "MasterFont")
	end
	
	scrollArea = Parrot_ScrollAreas:GetScrollArea(scrollArea)
	if not sticky then
		if not font then
			font = scrollArea.font or self.db.profile.font
		end
		if not fontSize then
			fontSize = scrollArea.fontSize or self.db.profile.fontSize
		end
		if not outline then
			outline = scrollArea.fontOutline or self.db.profile.fontOutline
		end
	else
		if not font then
			font = scrollArea.stickyFont or self.db.profile.stickyFont
		end
		if not fontSize then
			fontSize = scrollArea.stickyFontSize or self.db.profile.stickyFontSize
		end
		if not outline then
			outline = scrollArea.stickyFontOutline or self.db.profile.stickyFontOutline
		end
	end
	if outline == "NONE" then
		outline = ""
	end
	fs:SetFont(SharedMedia:Fetch('font', font), fontSize, outline)
	if color then
		local num = tonumber(color, 16)
		fs:SetTextColor(math.floor(num/256^2)/255, (math.floor(num/256)%256)/255, (num%256)/255)
	else
		fs:SetTextColor(1, 1, 1)
	end
	fs:SetText(text)
	fs.start = GetTime()
	fs.scrollArea = scrollArea
	fs.sticky = sticky
	local animationStyle
	if sticky then
		animationStyle = scrollArea.stickyAnimationStyle
	else
		animationStyle = scrollArea.animationStyle
	end
	local func, cleanup, overlap = Parrot_AnimationStyles:GetAnimationStyle(animationStyle)
	fs:SetPoint("BOTTOM", 0, 0)
	if not wildFontStrings then
		wildFontStrings = new()
		self:ScheduleRepeatingEvent("Parrot_OnUpdate", self.OnUpdate, 0, self)
	end
	local wildFontStrings_scrollArea = wildFontStrings[scrollArea]
	if not wildFontStrings_scrollArea then
		wildFontStrings_scrollArea = new()
		wildFontStrings[scrollArea] = wildFontStrings_scrollArea
	end
	local wildFontStrings_scrollArea_func = wildFontStrings_scrollArea[func]
	if not wildFontStrings_scrollArea_func then
		wildFontStrings_scrollArea_func = new()
		wildFontStrings_scrollArea[func] = wildFontStrings_scrollArea_func
		wildFontStrings_scrollArea_func.cleanup = cleanup
		wildFontStrings_scrollArea_func.overlap = overlap
	end	
	wildFontStrings_scrollArea_func.length = scrollArea[sticky and "stickySpeed" or "speed"] or 3
	local fontStringIDs_scrollArea = fontStringIDs[scrollArea]
	if not fontStringIDs_scrollArea then
		fontStringIDs_scrollArea = new()
		fontStringIDs[scrollArea] = fontStringIDs_scrollArea
	end
	local fontStringIDs_scrollArea_func = fontStringIDs_scrollArea[func]
	if not fontStringIDs_scrollArea_func then
		fontStringIDs_scrollArea_func = 1
	else
		fontStringIDs_scrollArea_func = fontStringIDs_scrollArea_func + 1
	end
	fontStringIDs_scrollArea[func] = fontStringIDs_scrollArea_func
	fs.id = fontStringIDs_scrollArea_func
	
	table.insert(wildFontStrings_scrollArea_func, 1, fs)
	fs:Show()
	self:OnUpdate(scrollArea, func)
end
Parrot.ShowMessage = Parrot_Display.ShowMessage

local function isOverlapping(alpha, bravo)
	if alpha:GetLeft() <= bravo:GetRight() and bravo:GetLeft() <= alpha:GetRight() and alpha:GetBottom() <= bravo:GetTop() and bravo:GetBottom() <= alpha:GetTop() then
		return true
	end
end

function Parrot_Display:OnUpdate()
	local now = GetTime()
	for scrollArea, u in pairs(wildFontStrings) do
		for animationStyle, t in pairs(u) do
			local t_len = #t
			local lastFS = nil
			for i, fs in ipairs(t) do
				local start, length = fs.start, t.length
				if start + length <= now then
					for j = i, t_len do
						local fs = t[j]
						local cleanup = t.cleanup
						if cleanup then
							cleanup(fs, scrollArea, fs.sticky)
						end
						fs:Hide()
						t[j] = nil
						freeFontStrings[fs] = true
					end
					break
				end
				local percent = (now - start) / length
				if percent >= 0.75 then
					fs:SetAlpha((1-percent) * 4)
				end
				
				fs:ClearAllPoints()
				animationStyle(fs, scrollArea, percent, fs.sticky, i, t_len, fs.id)
				
				if t.overlap and lastFS and isOverlapping(lastFS, fs) then
					local done = false
					local minimum = percent
					local maximum = 1
					local current = (percent + maximum) / 2
					while maximum - minimum > 0.01 do
						animationStyle(fs, scrollArea, current, fs.sticky, i, t_len, fs.id)
						if isOverlapping(lastFS, fs) then
							minimum = current
						else
							maximum = current
						end
						current = (maximum + minimum) / 2
					end
					current = current + 0.01
					fs.start = -current * length + now
					animationStyle(fs, scrollArea, current, fs.sticky, i, t_len, fs.id)
					for j = i+1, t_len do
						local v = t[j]
						if v.start > fs.start then
							v.start = fs.start
						end
					end
				end
				lastFS = fs
			end
			if not t[1] then
				u[animationStyle] = del(t)
			end
		end
		if not next(u) then
			wildFontStrings[scrollArea] = del(u)
		end
	end
	if not next(wildFontStrings) then
		wildFontStrings = del(wildFontStrings)
		self:CancelScheduledEvent("Parrot_OnUpdate")
	end
end
